# version 9.3.94
[CmdLetBinding()]
param(
	[string]$username,
	[string]$password,
	[Switch]$logToConsole,
	[string]$inputFile
)

try {
	cd $PSScriptRoot
	$logFileName = 'workflow_events.{0}.log' -f (get-date -Format 'yyyy-MM')
	. ./utils.ps1 -logFileName $logFileName -logToConsole:$logToConsole
	$api = [mm]::getAPI($username, $password)
} catch {
	Write-Output ('Got exception while initializing: ' + $_.Exception.GetBaseException().Message)
	exit 1
}

Enum Templates {
	processed
	rejected
	failed
	submitted
	approved
}

function sendEmail([string[]]$to, [Templates]$template, [hashtable]$extraTokens = $null, [switch]$includeState, $cc = $null){
	function getRequestDescription (){
		$changes = @()
		if ($objType -eq 'DNSRecord'){
			foreach ($change in $changeRequest.dnsRecordChanges.dnsRecordChange){
				$changes += $change
			}
			return [htmlUtils]::getRecordChangeListAsHTML($changes,$includeState)
		} else {
			throw "Don't know how to get request description for $objType requests"
		}
	}

	try {
		$tokens = @{
			requester = $requester
			requestID = $requestID
			requestType = $requestType
			application = $requestPropMap.Application
			group = $requestPropMap.Group
			requestComment = $changeRequest.saveComment
			scheduledDate = (?? $changeRequest.requestDate 'Immediately after approval')
			description = getRequestDescription
		}
		if ($null -ne $extraTokens){
			$tokens += $extraTokens
		}

		switch ($template){
			processed {
				$file    = "request-processed.htm"
				$subject = "Your Men&Mice $requestType Request has been implemented"
			}
			rejected {
				$file    = "request-rejected.htm"
				$subject = "Your Men&Mice $requestType Request has been rejected"
			}
			failed {
				$file    = "request-failed.htm"
				$subject = "Error(s) occurred while implementing Men&Mice $requestType Request #$requestID"
			}
			submitted {
				$file    = "request-submitted.htm"
				$subject = "A new Men&Mice $requestType Request is pending your review"
			}
			approved {
				$file    = "scheduled-request-approved.htm"
				$subject = "Your scheduled Men&Mice $requestType Request has been approved"
			}
		}

		if ($file -and $subject){
			[mmEmail]::prepareAndSend($to, $subject, $file, $cc, $mmConfig.email.bcc, $tokens)
		} else {
			log 'Unexpected Error: No template file or subject'
		}
	} catch {
		logException 'while sending email'
	}
}

function isFullyApproved () {
	$stateMap = @{}
	foreach ($change in (getFlatChanges)){
		if (notEmpty $change){
			$stateMap[$change.state] = $true
		}
	}

	foreach ($state in $stateMap.Keys){
		if ((notEmpty $state) -and $state -ne 'Approved'){
			logv2 ('Request is not fully approved')
			return $false
		}
	}

	logv2('Request is fully approved')
	return $true
}

function getFlatChanges () {
	$changes = @()
	# when a single change, this becomes
	$keys = @(
		'dnsRecordChange'
		#'dhcpReservationChange'
		#'dhcpScopeChange'
		#'dhcpAddressPoolChange'
		#'dhcpExclusionChange'
		#'dhcpOptionChange'
		#'dnsZoneChange'
		#'customPropertyChange'
		#'rangeChange'
	)
	foreach ($key in $keys){
		$s = $key+'s'
		if ($changeRequest.$s.$key){
			$changes += @($changeRequest.$s.$key) # if e.g. dnsRecordChanges.dnsRecordChange is an array it gets concatenated. If one object it gets added
		}
	}
	return $changes
}



function applyDNSRecordRequest () {

	function applyOwnership() {

		function setRecordOwnership($recordRef, [hashtable]$props) {
			log ("Setting ownership properties for $($recordRef): " + [mmJSON]::stringify($props))
			$api.setProperties($recordRef, 'DNSRecord', $props, $IDsaveComment)
		}

		$propsAdd = @{
			Owner = $requester
			Group = $requestPropMap.Group
			Application = $requestPropMap.Application
			Date = [mm]::getCurrentDay()
			Request = $requestID
		}
		$changes = [hashtable[]](getFlatChanges)
		foreach ($change in $changes){
			try {
				if ($change.state -ne 'Applied'){
					# don't register ownership on changes that aren't applied, likely failed
					# likely all of them are failed - so don't log anything
					continue
				}
				if ([mm]::isEmpty($change.dnsRecordAfter)){
					# only process modify or add
					continue
				}

				$ownModifiedRecords = ?? $Global:mmConfig.workflow.ownModifiedRecords 'orphaned'
				$rec = $change.dnsRecordAfter
				$recRef = ?? $rec.ref $rec.name # this will mean the records will be retrieved by ref after M&M v10
				$recName = $rec.name
				if ($change.type -eq 'ModifyDNSRecord'){

					$propsModify = @{
						Date = [mm]::getCurrentDay()
						Request = $requestID
					}
					if ([mm]::notEmpty($requestPropMap.Application)){
						$propsModify.Application = $requestPropMap.Application
					}

					if ($ownModifiedRecords -eq 'orphaned' -or $ownModifiedRecords -eq 'empty'){
						# we need to fetch the properties to know if it's orphaned
						$propMap = $api.getCustomProperties($recRef, 'DNSRecord')
						$owner = $propMap.Owner
						$group = $propMap.Group
						$requestGroup = $requestPropMap.Group

						if ([mm]::isEmpty($owner)){
							# empty-> definitely orphaned
							log "No owner registed on record $recName being modified"
							$propsModify.Owner = $requester
						} elseif ($owner -eq $requester){
							# nothing needs to be done here
							log "$requester is already the registered owner for record $recName that is being modified"
						} elseif ($ownModifiedRecords -eq 'orphaned' -and ![adUtils]::verifyUserByName($owner)){
							log "$owner is registered as owner for record $recName that is being modified, but $owner was not found in AD. Registering $requester as Owner"
							$propsModify.Owner = $requester
						}

						if ([mm]::isEmpty($group)){
							log "No group registered on record $recName being modified"
							$propsModify.Group = $requestGroup
						} elseif ($group -eq $requestGroup){
							log "Group $group is already registered as owning record"
						} elseif ($ownModifiedRecords -eq 'orphaned' -and ![adUtils]::verifyGroup($group)){
							log "Group $group is registered for record $recName that is being modified, but was not found in AD. Registering $requestGroup as Group for record"
							$propsModify.Group = $requestGroup
						}

						setRecordOwnership $recRef $propsModify
					} elseif ($ownModifiedRecords -eq 'always'){
						$propsModify.Owner = $requester
						$propsModify.Group = $requestPropMap.Group
						setRecordOwnership $recRef $propsModify
					}
				} elseif ($change.type -eq 'AddDNSRecord'){
					setRecordOwnership $recRef $propsAdd
				} else {
					log "AssertionError: Unexpected change type $($change.type) when dnsRecordAfter exists"
				}
			} catch {
				logException "while setting properties for record $recName"
			}
		}
	}

	function sendProcessedEmail(){
		# the processed email goes to the requester with the group on cc
		log "Sending processed email notification to requester $requester"
		$requesterEmail = [adUtils]::getEmailForUser($requester)
		$group = [adUtils]::getGroup($requestPropMap.Group)

		sendEmail -to @($requesterEmail) -cc $group.mail -template ([Templates]::processed) -includeState  # include the state just to highligh that they are "Applied"
	}

	function sendFailureEmail(){

		# start by getting the failure errors
		try {
			$events = $api.call('GetHistory', @{objRef = $requestRef; startDate = $theEvent.date; endDate = $theEvent.date; eventType = 'Error'}).events
		} catch {
			logException 'while retrieving error history'
		}

		$theErrors = @()
		foreach ($ev in $events){
			$theErrors += ($ev.eventText -replace '\[Request ID\: \d+\]$','').replace('Got an error while applying the change request:','')
		}

		if (!$theErrors.Count) {
			log ('Unable to find any appropriate Failure/Error messages. Using "Unknown error. Please inform dns@intel.com"')
			$theErrors = @('Unknown error.')
		}

		# send the failure notification to the requester and have the group on cc
		$group = [adUtils]::getGroup($requestPropMap.Group)

		$extraTokens = @{
			reason = [htmlUtils]::getArrayAsHTMListItems($theErrors)
		}
		log "Sending email to approver $approvers with requester $requester on cc"
		sendEmail -to @([adUtils]::getEmailForUser($requester)) -template ([Templates]::failed) -includeState -extraTokens $extraTokens -cc $group.mail
	}

	# apply ownership regardless of state - will only apply to changes in applied state
	applyOwnership

	if ($changeRequest.state -eq 'Applied'){
		sendProcessedEmail
	} elseif ($changeRequest.state -eq 'Failed'){
		sendFailureEmail
	} else {
		log "AssertionFailure: Unexpected request state for an applied request $($changeRequest.state)"
	}
}

function processRejectChanges () {
	try {
		$tokens = @{
			reason = $theEvent.comment
			rejecter = if ($isScriptRunner) {(?? $mmConfig.email.scriptUserFriendlyName 'Men&Mice')} else { $invokee }
		}

		$manager = [adUtils]::getManagerForUser($requester)

		log "Sending a rejection notification to $requester, $($manager.name) on cc"
		sendEmail -to @([adUtils]::getEmailForUser($requester)) -cc $manager.mail -template ([Templates]::rejected) -extraTokens $tokens -includeState
	} catch {
		logException 'while sending rejection notification'
	}
}

function processApproveChanges () {

	if (isFullyApproved){
		log ('Request is now fully approved.')
		# If something needs to be done after approval but before the changes are applied, do it here

		if ($isScriptRunner) {
			log 'Detected that request is being automated by workflow'
		}

		if ($isScheduled){
			log "Sending notification to $requester that scheduled request has been approved"
			sendEmail -to @([adUtils]::getEmailForUser($requester)) -template ([Templates]::approved) -includeState
		}

	} else {
		log ('Request *not* fully approved yet.')
	}
}

function processAddChangeRequest () {
	try {
		$manager = [adUtils]::getManagerForUser($requester)
		$groupInfo = [adUtils]::getGroup($requestPropMap.Group)
		if ($groupInfo -and $groupInfo.mail){
			log "Sending notification to group $($groupInfo.mail), $($manager.mail) on cc"
			sendEmail -to @($groupInfo.mail) -template ([Templates]::submitted) -cc $manager.mail
		} else {
			log ('Group {0} not found or mail not configured. Sending to manager {0} only.' -f $requestPropMap.Group, $manager.mail)
			sendEmail -to @($manager.mail) -template ([Templates]::submitted)
		}
	} catch {
		logException 'while processing AddChangeRequest'
	}
}

function processCheckApproveRejectChanges(){
	$orgAction = $eventAction.replace('ChangesCheck','') # => Approve or Reject
	if ($changeRequest.state -eq 'Failed') {
		log "Request is in Failed state. Anyone allowed to $orgAction request, to either retry or give up on request - allowed"
		return
	}

	# if user has access to all the changes, we assume it's an admin and simply allow it
	# if not, we check that the invokee is a member of the group registered on the request

	function hasAccessToAllPendingChanges () {
		foreach ($change in getFlatChanges){
			if ($change.state -eq 'Pending' -and [int]($change.hasAccess) -eq 0){
				return $false
			}
		}
		return $true
	}

	if (hasAccessToAllPendingChanges){
		log "$invokee has access to all changes in the request. Group membership not checked - allowed"
		return
	}

	$approvalGroup = $requestPropMap.Group

	if ([adUtils]::isUserMemberInGroup($invokee, $approvalGroup)){
		log "$invokee is a member in approval group $approvalGroup - allowed"
	} else {
		$msg = "You need to be a member of $approvalGroup to $orgAction this request"
		log "Returning error: $msg"
		[Console]::Out.Write($msg)
	}
}



try {
if ([mm]::notEmpty($inputFile)){
	logv2 "Reading input from $inputFile"
	$jsonStr = Get-Content -Raw $inputFile
} else {
	$jsonStr = [Console]::In.ReadToEnd()
	logv2 $jsonStr
}

$jsonInput = [mmJSON]::parse($jsonStr)
$inputParams = $jsonInput.externalScriptParameters
$theEvent = $inputParams.event
$eventAction = $theEvent.action
if ([mm]::isEmpty($inputFile)){
	[mm]::storeDebugBlob($jsonStr, $eventAction, 'json')
}
$changeRequest = ?? $inputParams.requestAfter.changeRequest $inputParams.requestBefore.changeRequest
$requestPropMap = [mm]::getPropertyMap(@($changeRequest.customProperties.property))
$objType = $changeRequest.objType
$requestRef = $changeRequest.ref
$requestID = [mm]::ref2ID($requestRef)
$requestType = 'DNS'

$requester = [mm]::cleanXmlString($changeRequest.requester)
$invokee = [mm]::cleanXmlString($theEvent.userName)
$isScriptRunner = $invokee -eq 'ScriptRunner'
$isScheduled = [mm]::notEmpty($changeRequest.requestDate)
$IDsaveComment = "RequestID:$requestID"



log ('Processing #{0}: {1}. Requested by: {2}. Command invoked by: {3}' -f $requestID,$eventAction,$requester,$invokee)

try {
	switch ($eventAction) {
		'applyDNSRecordChanges' { applyDNSRecordRequest ; break}
		'RejectChanges' { processRejectChanges ; break }
		'RejectDNSRecordChanges' { processRejectChanges ; break }
		'ApproveChanges' { processApproveChanges ; break }
		'ApproveDNSRecordChanges' { processApproveChanges ; break }
		'AddChangeRequest' { processAddChangeRequest ; break }
		'AddDNSRecordChangeRequest' { processAddChangeRequest ; break }
		'ApproveChangesCheck' { processCheckApproveRejectChanges; break }
		'RejectChangesCheck' { processCheckApproveRejectChanges; break }
		default {
			log ('ERROR: Unsupported eventAction "{0}" ' -f $eventAction)
			exit 1
		}
	}
} catch {
	logException 'while processing request event. Notify M&M'
}

log 'Done'
$api.logout()

} catch {
	logException 'during processing'
	$api.logout()
}
