﻿# version 9.3.94
[CmdLetBinding()]
param(
	$username,
	$password,
	[switch]$logToConsole,
	[switch]$forceProcessing,
	[string]$inputFile
)

try {
	cd $PSScriptRoot
	$logFileName = 'workflow_submit.{0}.log' -f (get-date -Format 'yyyy-MM')
	. ./utils.ps1 -logFileName $logFileName -logToConsole:$logToConsole
	$api = [mm]::getAPI($username, $password)
} catch {
	Write-Output ('Got exception while initializing: ' + $_.Exception.GetBaseException().Message)
	exit 1
}


function verifyChanges([int]$requestID, [string]$requester, [string]$group){

	$requestRef = [mm]::ID2ObjRef($requestID, 28)
	log 'Fetching request'
	$changeRequest = $api.call('GetChangeRequest', @{changeRequestRef = $requestRef}).changeRequest
	log "Verifying $($changeRequest.dnsRecordChanges.Count) DNS Record changes"
	foreach ($change in $changeRequest.dnsRecordChanges){
		$changeType = $change.type.replace('DNSRecord','')
		$before = $change.dnsRecordBefore
		if ($changeType -eq 'Add'){
			$zonename = $api.call('GetDNSZone', @{dnsZoneRef = $change.dnsRecordAfter.dnsZoneRef}).dnsZone.name
			if ($change.dnsRecordAfter.name -eq $zonename){
				returnErrorToCentral "$zonename is now an allowed record name: It is an apex record for the zone $zonename"
				exit
			}

			$pqdn = $change.dnsRecordAfter.name.replace('.'+ $zonename, '')
			if ($pqdn.split('.').Count -gt 1){
				returnErrorToCentral "$pqdn is not an allowed hostname: Only single label hostnames are allowed."
				exit
			}
			continue # don't need to validate ownership for Adds
		}

		$ref = ?? $before.ref $before.name
		$recName = $before.name
		$propMap = $api.getCustomProperties($ref, 'DNSRecord')
		$recOwner = $propMap.Owner
		$recGroup = $propMap.Group
		if ($recOwner -or $recGroup){
			# user can request changes to the record if he's the owner or is a member of the group registered to the record
			# or if the registered group no longer exists
			log "$requester ($group) is requesting $changeType record $recName that is registered to $recOwner ($recGroup)"

			if ($recOwner -ne $owner -and $recGroup -ne $group){
				# requester is neither the owner nor is requesting on behalf of the registered group - we need to verify!

				log "Verifying that $requester is a member of group $recGroup"
				$groupInfo = [adUtils]::getGroup($recGroup)
				if ($groupInfo){
					if (![adUtils]::isUserMemberInGroupByDN($requester, $groupInfo.dn)){
						log "$requester is not a member in the group $recGroup. $changeType request not allowed"
						return "You cannot request to $changeType the record $recName. It is registered to the group $recGroup, which you are not a member of."
					} else {
						log "$requester is a member! Request allowed"
					}
				} else {
					log "Group `"$group`" was not found in AD. $changeType request allowed"
				}
			}
		} else {
			log "No ownership found on record $recName. $changeType request allowed"
		}
	}
	return ""
}


$newProperties = @{}

try {
	if ([mm]::notEmpty($inputFile)){
		logv2 "Reading input from $inputFile"
		$xmlStr = Get-Content -Raw $inputFile
	} else {
		$xmlStr = [Console]::In.ReadToEnd()
		[mm]::storeDebugBlob($xmlStr, 'submit', 'xml')
		logv2 ("Input:" + $xmlStr)
	}


	$xml = [xml]($xmlStr)
	$requester = [mm]::cleanXmlString(($xml.selectSingleNode("externalScriptParameters")).getAttribute('userName'))
	$cfs = $xml.selectSingleNode("externalScriptParameters/customFields")
	$groupStr = [mm]::cleanXmlString($cfs.selectSingleNode("customField[@customFieldName='Group']").value)
	$ApproversProp = [mm]::cleanXmlString($cfs.selectSingleNode("customField[@customFieldName='Approvers']").value)
	$object = $xml.externalScriptParameters.object
	$requestID = [int]$object.id

	log ("Processing request {0} submitted by {1}" -f $requestID, $requester)

	# new requests don't have "Approvers" property set - should be read only field on ChangeRequests
	if ([mm]::notEmpty($ApproversProp) -and !$forceProcessing){
		log ('Request already processed.')
		exit 0
	}

	$adGroup = [adUtils]::getGroup($groupStr)
	if (!$adGroup){
		returnErrorToCentral "No AD group found that matches `"$groupStr`""
		exit
	}
	if (!$adGroup.mail){
		returnErrorToCentral "A group was found in AD that matches `"$groupStr`", but it is not configured with an email address. Please enter a distribution group with email configured."
		exit
	}
	if (![adUtils]::isUserMemberInGroupByDN($requester, $adGroup.dn)){
		returnErrorToCentral "You are not a member in the AD group `"$groupStr`". Please enter a group that you are a member of."
	}

	if ($adGroup.samAccountName -ne $groupStr){
		$newProperties.Group = $adGroup.samAccountName
	}

	# start by adding the members of the group as approvers
	$approvers = @()
	$groupMembers = [adUtils]::getUsersInGroupByDN($adGroup.dn, @(), $true)
	foreach ($member in $groupMembers){
		$memberName = $member.'msDS-PrincipalName'
		if ($memberName -ne $requester -and $memberName -notin $approvers){
			$approvers += $memberName
		}
	}

	if (!$approvers.Count){
		returnErrorToCentral "The group you selected has too few users as members"
	}

	# add the manager if not already in the approvers list
	$manager = [adUtils]::getManagerForUser($requester)
	if ($manager -and $manager.name){
		if ($manager.name -notin $approvers){
			log ('Manager {0} not part of group. Adding to approvers list' -f $manager.name)
			$approvers += $manager.name
		}
	} else {
		log ('No manager registered for user in AD!')
	}

	# If there are that many approvers that we can't save it in the database, just store the first 890 characters.
	# That should be enough for 100 approvers or so
	$approversStr = $approvers -join "`n"
	if ($approversStr.Length -gt 890){
		$approversStr = $approversStr.Substring(0,890)
	}

	$newProperties.Approvers = $approversStr

	if ($Global:mmConfig.workflow.verifyExistingRecordOwnership -ne $false){
		try {
			$verificationError = verifyChanges -requestID $requestID -requester $requester -group $adGroup.samAccountName
			if ($verificationError){
				returnErrorToCentral $verificationError
			}
		} catch {
			logException 'while verifying changes'
			# don't fail the request even if there's an unexpected error while verifying the changes, such as API error, since that shouldn't happen
		}
	}

	returnSuccessToCentral $newProperties
} catch {
	logException "while processing input xml"
	log $xmlStr # repeated if in higher verbosity
	returnErrorToCentral "Unhandled exception"
}



